% :- type mezo      --->  0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9.
% :- type sor         ==  list(mezo).
% :- type ujjtabla    ==  list(sor).

% :- type ujj       --->  s | sw | w | nw | n | ne | e | se.
% :- type ujjak       ==  list(ujj).
% :- type peremezes --->  u(ujjak, ujjak, ujjak, ujjak).

% :- type file        ==  atom.

% :- pred ujjak(ujjtabla::in, peremezes::out).

:- module(ujjak_keret,
	  [
	   helyes_feladvany/1,  % helyes_feladvany(ujjtabla::in)
	   ujjak_be/2,		% ujjak_be(file::in, ujjtabla::out).
	   ujjak_ki/2,		% ujjak_ki(file::in, ujjtabla::in, peremezes::in).
	   megold/2,		% megold(file::in, file::in)
	   stopper/2		% stopper(file::in, file::in)
	  ]
	 ).

:- use_module(library(lists), [member/2,non_member/2,
			       remove_duplicates/2,last/2]).

%-------------- Bejaratok ------------------------------------

helyes_feladvany(Tabla) :-
	%% negyzet alaku
	length(Tabla, N),
	all_same_length(Tabla, N),
	%% nem ures
	N \= 0,
	%% nincs nem helyes eleme a Tabla ujjtablanak
	\+ (   member(Sor, Tabla),
	       member(Mezo, Sor),
	       \+ helyes_mezo(Mezo)
	   ).

% helyes_mezo(+Mezo): Mezo egy helyes mezot ir le
helyes_mezo(D) :- number(D), 0 =< D, D =< 9.

% ujjak_be(+File, -Tabla):
% A File-bol beolvashato a Tabla altal megadott feladat.
ujjak_be(File, Tabla) :-
	open_file(File, read, Stream),
	call_cleanup(read_in(Stream, Characters),
		     close(Stream)),
	parse_data(Characters, Tabla).

% ujjak_ki(+File, +Tabla, +Peremezes):
% A Tabla feladat Peremezes megoldasat kiirja a File allomanyba.
ujjak_ki(File, Tabla, Peremezes) :-
	open_file(File, write, Stream),
	current_output(OOut),
	set_output(Stream),
	call_cleanup(pretty_print(Tabla, Peremezes),
		     (set_output(OOut),close(Stream))).

% megold(+FileIn, +FileOut):
% A FileIn allomanybol beolvasott feladat minden megoldasat
% kiirja a FileOut allomanyba.
megold(FileIn, FileOut) :-
	ujjak_be(FileIn, Tabla),
	ensure_loaded(user:ujjak),
	open_file(FileOut, write, Stream),
	current_output(OOut),
	set_output(Stream),
	call_cleanup(print_solutions(Tabla),
		     (set_output(OOut),close(Stream))).

% stopper(+FileIn, +FileOut):
% Mint megold/2, de a futashoz szukseges idot is kiirja.
stopper(FileIn, FileOut) :-
	ensure_loaded(user:ujjak), % hogy ne a mert idobol menjen
	statistics(runtime, [T0|_]),
	megold(FileIn, FileOut),
	statistics(runtime, [T1|_]),
	T is T1 - T0,
	format('Feladat: ~a, futsi id: ~3d s~n', [FileIn,T]).

%-------------- Allomanyok kezelese --------------------------

% open_file(+File, +Mode, -Stream):
% A File allomanyt megnyitja a Stream folyam formajaban Mode modban.
open_file(user, M, S) :-
	!, standard_stream(M, S).
open_file(File, Mode, Stream) :-
	open(File, Mode, Stream).

% standard_stream(?Mode, ?Stream):
% A Mode modu szabvanyos folyam a Stream.
standard_stream(read, user_input).
standard_stream(write, user_output).

%-------------- Beolvasas ------------------------------------

% read_in(+S, -L):
% S folyambol beolvashato karakterek listaja L.
read_in(S, L) :-
	get_code(S, C),
	read_in(C, S, L).

% read_in(+C, +S, -L):
% Ha C az S folyambol elozoleg beolvasott karakter, akkor L az S-bol
% beolvashato maradek karakterek listaja, C-t is beleertve.  Ha C = -1,
% akkor L = [].
read_in(C, _, L) :-
	C =:= -1, !, L = [].
read_in(C, S, [C|L]) :-
	get_code(S, C1),
	read_in(C1, S, L).

% parse_data(+Chars, -Table):
% Chars fuzerbol kielemezheto egy Tabla tabla
parse_data(Chars, Table) :-
	parse_table(Table, Chars, Rest), !,
	correct_syntax(Rest),
	correct_table(Table).
parse_data(_, _) :-
	warning('A bemeneti llomny nem megfelel formj.',[]),
	fail.

%-------------- Ellenorzesek ---------------------------------

% correct_syntax(+Rest):
% A Rest fuzer a bemenet helyes maradeka, azaz ures.
correct_syntax(Rest) :-
	Rest = [_|_], !, 
	chars_of_type(mid_line, L, Rest, _),
	warning('Nem vrt karakter(ek): "~s"', [L]),
	fail.
correct_syntax(_).

% correct_table(+Table):
% A Table altal megadott tabla szintaktikailag helyes.
correct_table(Table) :-
	%% negyzet alaku
	(   length(Table, N), all_same_length(Table, N)
	->  true
	;   warning('A tbla nem ngyzet alak.', []), fail
	),
	%% nem ures
	(   N = 0
	->  warning('A tblnak nincsenek mezi.', []), fail
	;   true
	).

%-------------- DCG elemzes ----------------------------------

% Kielemezheto egy tabla
parse_table([Row|Rows]) -->
	table_row(Row),
	newline, !,
	parse_table(Rows).
parse_table([]) -->
	chars_of_type(layout, _).

% Kielemezheto a tabla egy sora
table_row([V|Vs]) -->
	nat_number(V), {0 =< V, V =< 9}, !,
	table_row(Vs).
table_row([]) --> [].

% Kielemezheto egy (tizes szamrendszerbeli) Int termeszetes szam, amit
% esetleg szokozok eloznek meg.
nat_number(Int) -->
	chars_of_type(horizontal, _),
	chars_of_type(digit, Ds),
	{Ds = [_|_], number_codes(Int, Ds)}.

newline -->
	chars_of_type(horizontal, _),
	char_of_type(vertical, _).

% layout_char(Dir, C):
% horizontal_layout_char(C): C  egy vizszintes iranyu
% nem lathato karakter.
horizontal_layout_char(0' ).
horizontal_layout_char(0'\t).
horizontal_layout_char(0'\r).

% vertical_layout_char(C): C  egy fuggoleges iranyu
% nem lathato karakter.
vertical_layout_char(0'\n).
vertical_layout_char(0'\f).
vertical_layout_char(0'\v).

% Kielemezheto egy Type tipusu C karakter.
char_of_type(Type, C) -->
	[C], {char_type(Type, C)}.

% chars_of_type(Type, Cs) -->
% Kielemezheto Type tipusu karakterek egy Cs listaja.	
chars_of_type(Type, [C|Cs]) -->
	char_of_type(Type, C), !,
	chars_of_type(Type, Cs).
chars_of_type(_, []) --> [].

% char_type(Type, C) :
% A C karakter Type tipusu.
char_type(layout, C) :-
	(   vertical_layout_char(C) -> true
	;   horizontal_layout_char(C)
	).
char_type(digit, D) :-
	D >= 0'0, D =< 0'9.
char_type(mid_line, C) :-
	\+ vertical_layout_char(C).
char_type(horizontal, C) :-
	horizontal_layout_char(C).
char_type(vertical, C) :-
	vertical_layout_char(C).

%-------------- Kiiratas -------------------------------------

% print_solutions(+Tabla):
% A Tabla altal megadott feladat osszes megoldasat
% kiirja az aktualis kimenetre.
print_solutions(Tabla) :-
	user:ujjak(Tabla, Peremezes),
	pretty_print(Tabla, Peremezes),
	fail.
print_solutions(_).

% pretty_print(Table)
% Kiirja a Table tablat ujjak nelkul.
pretty_print(Table) :-
	pretty_print(Table, _), !.

% pretty_print(Table, Arrows)
% Kiirja a Table tabla megoldasait Arrows peremezessel
pretty_print(Table, u(N,E,S,W)) :-
	well_formed_solution(Table, N,E,S,W),
	print(' '), print_arrow_line(N),
	Table = [Row|_],
	print('  '), print_separator(Row),
	print_lines(Table, W, E),
	print(' '), print_arrow_line(S), nl.

% well_formed_solution(+Table, +N, +E, +S, +W)
% Ellenorzi, hogy a megoldas megfelel-e a formai kovetelmenyeknek
well_formed_solution(Table, N,E,S,W) :-
	(   length(Table, Len),
	    all_same_length([N,E,S,W], Len)
	->  true
	;   warning('Hibs mret megolds!', []), fail
	),
	(   contains_only(N, [sw,s,se]),
	    contains_only(E, [nw,w,sw]),
	    contains_only(S, [nw,n,ne]),
	    contains_only(W, [ne,e,se]),
	    \+ first(N, sw), \+ last(N, se),
	    \+ first(E, nw), \+ last(E, sw),
	    \+ first(S, nw), \+ last(S, ne),
	    \+ first(W, ne), \+ last(W, se)
	->  true
	;   warning('A megolds nem felel meg a specifikcinak!', []), fail
	).

% print_arrow_line(Arrows)
% Kiirja az also vagy a felso ujjsort.
print_arrow_line([]) :- nl.
print_arrow_line([A|As]) :-
	print('   '), print_arrow(A),
	print_arrow_line(As).

% print_lines(Rows, W, E)
% Kiirja a Rows tabla sorait W ujjakkal elottuk, E ujjakkal utanuk.
print_lines([], [], []).
print_lines([R|Rs], [W|Ws], [E|Es]) :-
	print_arrow(W), print(' '),
	print_line(R),
	print('| '), print_arrow(E), nl,
	print('  '), print_separator(R),
	print_lines(Rs, Ws, Es).

% print_line(Row)
% Kiirja a tabla Row sorat.
print_line([]).
print_line([F|Fs]) :-
	format('| ~p ', [F]),
	print_line(Fs).

% print_separator(R)
% Kiir egy elvalaszto sort az R sor ala/fole.
print_separator([]) :-
	print('+'), nl.
print_separator([_|R]) :-
	print('+---'),
	print_separator(R).

% print_arrow(A)
% Kiirja az A ujjhoz tartozo karaktert.
print_arrow(A) :-
	arrow_char(A,C),
	put_code(C).

% arrow_char(A, C)
% Az A ujj a C karakterrel jelenitheto meg ASCII terminalon.
arrow_char(?,  0' ).
arrow_char(s,  0'|).
arrow_char(sw, 0'/).
arrow_char(w,  0'-).
arrow_char(nw, 0'\\).
arrow_char(n,  0'|).
arrow_char(ne, 0'/).
arrow_char(e,  0'-).
arrow_char(se, 0'\\).

%-------------- Seged-eljarasok ------------------------------

% all_same_length(Ls, L)
% Az Ls elemei listak, hosszuk azonosan L.
all_same_length([], _).
all_same_length([L|Ls], N) :-
	length(L, N),
	all_same_length(Ls, N).

% contains_only(+L, +Es)
% Az L lista csak az Es lista elemeibol all.
contains_only(L, Es) :-
	remove_duplicates(L, Xs),
	\+ (member(X, Xs), non_member(X, Es)).

% first(?List, ?First)
% List elso eleme First.
first([H|_], H).

% warning(+Txt, +Args):
% Kiirja a Txt szoveggel es Args argumentumlistaval
% megadott figyelmezteto szoveget.
warning(Txt, Args) :-
	print_message(warning, format(Txt, Args)).

%-------------- ISO kompatibilitas ---------------------------

% Az alabbi cel az ISO Prolog szabvanynak meg nem megfelelo, de
% DEC10 kompatiblis Prolog rendszerekben valo futtatashoz szukseges.

:- (   predicate_property(get_code(_,_), _) -> true
   ;   assert((get_code(S, C) :- get0(S, C))),
       assert((put_code(C) :- put(C))),
       assert((number_codes(N, C) :- number_chars(N, C)))
   ).
